@@ -7,25 +7,22 @@ import time
 from datetime import datetime
 
 import runez
-from runez.pyenv import pyenv_scanner, PythonDepot, Version
-
-from pickley.pypi import PypiInfo
+from runez.pyenv import PypiStd, PythonDepot, PythonInstallationScanner, Version
 
 
-__version__ = "2.5.5"
+__version__ = "3.3.0"
 LOG = logging.getLogger(__name__)
 PICKLEY = "pickley"
 DOT_META = ".pickley"
 K_CLI = {"delivery", "index", "python"}
 K_DIRECTIVES = {"include"}
 K_GROUPS = {"bundle", "pinned"}
-K_LEAVES = {"facultative", "install_timeout", "pyenv", "version", "version_check_delay"}
+K_LEAVES = {
+    "facultative", "install_timeout", "min_python", "preferred_min_python", "preferred_pythons", "pyenv", "version", "version_check_delay"
+}
 PLATFORM = platform.system().lower()
 
 DEFAULT_PYPI = "https://pypi.org/simple"
-DEFAULT_PYTHONS = "python3, python"
-RE_PYPI_CANONICAL = re.compile(r"^[a-z][a-z0-9-]*[a-z0-9]$")
-RE_PYPI_ACCEPTABLE = re.compile(r"^[a-z][a-z0-9._-]*[a-z0-9]$", re.IGNORECASE)
 
 
 def abort(message):
@@ -35,17 +32,6 @@ def abort(message):
     sys.exit(1)
 
 
-def canonical_pypi_name(original):
-    """
-    Args:
-        original (str): Name as given in setup.py
-
-    Returns:
-        (str): Corresponding canonical pypi name
-    """
-    return original and original.lower().replace("_", "-").replace(".", "-")
-
-
 def inform(message):
     """
     Args:
@@ -68,10 +54,10 @@ def despecced(text):
     if text and "==" in text:
         text = text.strip()
         i = text.index("==")
-        version = text[i + 2:].strip()
+        version = text[i + 2:].strip() or None
         text = text[:i].strip()
 
-    return text, version or None
+    return text, version
 
 
 def specced(name, version):
@@ -85,40 +71,74 @@ def specced(name, version):
     """
     name = name.strip()
     if version and version.strip():
-        return "%s==%s" % (name, version.strip())
+        return f"{name}=={version.strip()}"
 
     return name
 
 
 def pypi_name_problem(name):
-    if not name or not RE_PYPI_ACCEPTABLE.match(name):
+    if not PypiStd.is_acceptable(name):
         note = None
-        problem = "'%s' is not a valid pypi package name" % runez.red(name)
+        problem = f"'{runez.red(name)}' is not a valid pypi package name"
         if name and not name[0].isalpha():
             note = "\npickley intentionally refuses to look at names that don't start with a letter"
 
         if note:
             note += "\nIf you think this name is legit, please submit an issue https://github.com/codrsquad/pickley/issues"
-            problem = "%s\n%s" % (problem, note)
+            problem = runez.joined(problem, note, delimiter="\n")
 
         return problem
 
 
-def validate_pypi_name(name):
-    problem = pypi_name_problem(name)
-    if problem:
-        abort(problem)
+def _dynamic_resolver(cfg, name_or_url):
+    package_name, package_version = despecced(name_or_url)
+    if package_version:
+        return package_name, package_version, None
 
-    if not RE_PYPI_CANONICAL.match(name):
-        logging.warning("'%s' is not pypi canonical, use dashes only, and lowercase" % name)
+    cached_resolved = None
+    folder = None
+    is_git = "://" in name_or_url or name_or_url.endswith(".git")
+    if is_git or "/" in name_or_url:
+        safe_name = re.sub(r"\W+", "-", name_or_url).strip("-")
+        folder = name_or_url if not is_git else cfg.cache.full_path("checkout", safe_name)
+        cached_resolved = cfg.cache.full_path(f"{safe_name}.rlv")
+        info = runez.read_json(cached_resolved)
+        if info and "resolved" in info:
+            package_name, package_version = info["resolved"]
+            return package_name, package_version, folder
 
+        if is_git and not os.path.isdir(folder):
+            runez.ensure_folder(folder, clean=True)
+            if runez.DRYRUN:
+                package_name = os.path.basename(name_or_url)
+                if package_name.endswith(".git"):
+                    package_name = package_name[:-4]
 
-def git_clone(pspec):
-    basename = runez.basename(pspec.original)
-    pspec.folder = pspec.cfg.cache.full_path("checkout", basename)
-    runez.ensure_folder(pspec.folder, clean=True, dryrun=False)
-    logger = LOG.info if runez.DRYRUN else runez.UNSET
-    runez.run("git", "clone", pspec.original, pspec.folder, dryrun=False, logger=logger)
+                return package_name, "0.0.0", folder
+
+            runez.run("git", "clone", name_or_url, folder)
+
+    if folder:
+        setup_py = os.path.join(folder, "setup.py")
+        if not os.path.exists(setup_py):
+            abort(f"No setup.py in '{runez.red(runez.short(folder))}'")
+
+        with runez.CurrentFolder(folder):
+            # Some setup.py's assume current folder is the one with their setup.py
+            r = runez.run(sys.executable, "setup.py", "--name", dryrun=False, fatal=False, logger=False)
+            package_name = r.output
+            if r.failed or not package_name:
+                abort(f"Could not determine package name from '{runez.red(runez.short(setup_py))}'")
+
+            r = runez.run(sys.executable, "setup.py", "--version", dryrun=False, fatal=False, logger=False)
+            package_version = r.output
+            if r.failed or not package_version:
+                abort("Could not determine package version from setup.py")
+
+            runez.save_json(dict(resolved=[package_name, package_version]), cached_resolved)
+            return package_name, package_version, folder
+
+    return package_name, package_version, None
 
 
 class PackageSpec(object):
@@ -132,131 +152,139 @@ class PackageSpec(object):
 
     # The following fields become available after a call to resolve()
     name = None  # type: str
-    version = None  # type: str
     folder = None  # type: str
     dashed = None  # type: str
     wheelified = None  # type: str
     python = None  # type: runez.pyenv.PythonInstallation
     pinned = None  # type: str
     settings = None  # type: TrackedSettings
+    _desired_track = None  # type: TrackedVersion
+    _latest = None  # type: TrackedVersion
+    _manifest = None  # type: TrackedManifest
 
-    def __init__(self, cfg, name_or_url, version=None):
+    def __init__(self, cfg, name_or_url):
         """
         Args:
             cfg (PickleyConfig): Associated configuration
             name_or_url (str): Provided package reference (either name, folder or git url)
-            version (str | None): Optional version spec
         """
         self.cfg = cfg
         self.original = name_or_url
-        self.version = version
-        self._delayed_resolver = self._get_delayed_resolver()
-        if self._delayed_resolver is None:
-            self.resolve()
+        self.name, self.given_version, self.folder = _dynamic_resolver(cfg, name_or_url)
+        runez.abort_if(pypi_name_problem(self.name))
+        self.dashed = PypiStd.std_package_name(self.name)
+        if self.name != self.dashed:
+            logging.warning(f"'{runez.red(self.name)}' is not pypi canonical, use dashes only and lowercase")
+
+        self.wheelified = PypiStd.std_wheel_basename(self.name)
 
     def __repr__(self):
-        return self.specced or self.dashed or self.original
+        return specced(self.dashed, self.given_version)
 
     def __lt__(self, other):
         return str(self) < str(other)
 
-    def _get_delayed_resolver(self):
-        if "://" in self.original or self.original.endswith(".git"):
-            return git_clone
+    @runez.cached_property
+    def python(self):
+        return self.cfg.find_python(self)
 
-        if "/" in self.original:
-            folder = runez.resolved_path(self.original)
-            if os.path.isdir(folder):
-                self.folder = folder
-
-    def resolve(self):
-        if self.name:
-            return
-
-        if self._delayed_resolver is not None:
-            self._delayed_resolver(self)
-            self._delayed_resolver = None
-
-        self.name, self.version = self.resolved_name_version()
-        validate_pypi_name(self.name)
-        self.dashed = canonical_pypi_name(self.name)
-        self.wheelified = self.name.replace("-", "_").replace(".", "_")
-        self.pinned = self.cfg.pinned_version(self)
-        self.python = self.cfg.find_python(self)
-        self.settings = TrackedSettings(
+    @runez.cached_property
+    def settings(self):
+        return TrackedSettings(
             delivery=self.cfg.delivery_method(self),
             index=self.cfg.index(self) or self.cfg.default_index,
             python=self.python.executable,
+            virtualenv=self.cfg.get_virtualenv(self),
         )
 
-    def resolved_name_version(self):
-        if not self.folder:
-            if self.version:
-                return self.original, self.version
+    @runez.cached_property
+    def pinned(self):
+        return self.cfg.pinned_version(self)
 
-            return despecced(self.original)
+    @runez.cached_property
+    def _pickley_dev_mode(self):
+        """Path to pickley source folder, if we're running in dev mode"""
+        if self.dashed == PICKLEY:
+            dev_path = self.cfg.pickley_dev_path()
+            if dev_path:
+                return dev_path
 
-        setup_py = os.path.join(self.folder, "setup.py")
-        if not os.path.exists(setup_py):
-            abort("No setup.py in %s" % self.folder)
-
-        with runez.CurrentFolder(self.folder):
-            # Some setup.py's assume current folder is the one with their setup.py
-            r = runez.run(sys.executable, "setup.py", "--name", dryrun=False, fatal=False, logger=False)
-            package_name = r.output
-            if r.failed or not package_name:
-                abort("Could not determine package name from %s" % setup_py)
-
-            validate_pypi_name(package_name)
-            r = runez.run(sys.executable, "setup.py", "--version", dryrun=False, fatal=False, logger=False)
-            package_version = r.output
-            if r.failed or not package_version:
-                abort("Could not determine package version from setup.py")
-
-        return package_name, package_version
-
-    @property
-    def specced(self):
-        if self.name and self.version:
-            return specced(self.dashed, self.version)
-
-    @property
+    @runez.cached_property
     def index(self):
         """Index to use for this package spec"""
         return self.cfg.index(self)
 
     @property
-    def install_path(self):
-        if self.name:
-            if self.dashed == PICKLEY and runez.SYS_INFO.dev_folder():
-                return self.cfg.meta.full_path(PICKLEY, "%s-dev" % PICKLEY)
+    def desired_track(self):
+        if self._desired_track is None:
+            if self.given_version:
+                self._desired_track = TrackedVersion(source="explicit", version=self.given_version)
 
-            if self.version:
-                return self.cfg.meta.full_path(self.dashed, "%s-%s" % (self.dashed, self.version))
+            elif self.pinned:
+                self._desired_track = TrackedVersion(source="pinned", version=self.pinned)
+
+            else:
+                self._desired_track = self.get_latest()  # By default, the latest is desired
+                candidates = []
+                if self.is_currently_installed:
+                    candidates.append(TrackedVersion.from_manifest(self.manifest))
+
+                if self.dashed == PICKLEY:
+                    candidates.append(TrackedVersion(source="current", version=__version__))
+
+                for candidate in candidates:
+                    if candidate and candidate.version:
+                        dt = self._desired_track
+                        if not dt or not dt.version or Version(candidate.version) > Version(dt.version):
+                            self._desired_track = candidate
+
+        return self._desired_track
 
     @property
+    def is_up_to_date(self):
+        manifest = self.manifest
+        return manifest and manifest.version == self.desired_track.version and self.is_healthily_installed()
+
+    @property
+    def manifest(self):
+        """TrackedManifest: Manifest of the current installation of this package"""
+        if self._manifest is None:
+            self._manifest = TrackedManifest.from_file(self.manifest_path)
+
+        return self._manifest
+
+    @runez.cached_property
     def manifest_path(self):
-        return self.name and self.cfg.meta.full_path(self.dashed, ".manifest.json")
+        return self.cfg.meta.full_path(self.dashed, ".manifest.json")
 
-    @property
+    @runez.cached_property
     def meta_path(self):
-        return self.name and self.cfg.meta.full_path(self.dashed)
+        return self.cfg.meta.full_path(self.dashed)
 
-    @property
+    @runez.cached_property
     def ping_path(self):
         """Path to .ping file (for throttle auto-upgrade checks)"""
-        return self.name and self.cfg.cache.full_path("%s.ping" % self.dashed)
+        return self.cfg.cache.full_path(f"{self.dashed}.ping")
 
     @property
+    def is_currently_installed(self):
+        manifest = self.manifest
+        return manifest and manifest.version
+
+    @runez.cached_property
     def is_already_installed_by_pickley(self):
         """bool: True if this package was already installed by pickley once"""
-        if self.name:
-            return self.dashed == PICKLEY or os.path.exists(self.manifest_path)
+        return self.dashed == PICKLEY or os.path.exists(self.manifest_path)
+
+    def get_install_path(self, version):
+        if self._pickley_dev_mode:
+            return self.cfg.meta.full_path(PICKLEY, f"{PICKLEY}-dev")
+
+        return self.cfg.meta.full_path(self.dashed, f"{self.dashed}-{version}")
 
     def get_lock_path(self):
         """str: Path to lock file used during installation for this package"""
-        name = self.dashed or os.path.basename(self.original)
-        return self.cfg.meta.full_path("%s.lock" % name)
+        return self.cfg.meta.full_path(f"{self.dashed}.lock")
 
     def skip_reason(self, force=False):
         """str: Reason for skipping installation, when applicable"""
@@ -284,7 +312,7 @@ class PackageSpec(object):
             if os.path.getsize(target) == 0 or not runez.is_executable(target):
                 return True  # Empty file or not executable
 
-        for line in runez.readlines(target, default=[], first=5, errors="ignore"):
+        for line in runez.readlines(target, first=5):
             if PICKLEY in line:
                 return True  # Pickley wrapper
 
@@ -293,18 +321,27 @@ class PackageSpec(object):
 
     def is_healthily_installed(self):
         """Double-check that current venv is still usable"""
-        py_path = self.cfg.available_pythons.resolved_python_exe(self.install_path)
-        if py_path:
-            return runez.run(py_path, "--version", dryrun=False, fatal=False, logger=False).succeeded
+        manifest = self.manifest
+        if manifest and manifest.version:
+            if manifest.entrypoints:
+                for name in manifest.entrypoints:
+                    exe_path = self.exe_path(name)
+                    if not runez.is_executable(exe_path):
+                        return False
+
+            install_folder = self.get_install_path(manifest.version)
+            py_path = self.cfg.available_pythons.resolved_python_exe(install_folder)
+            if py_path:
+                return runez.run(py_path, "--version", dryrun=False, fatal=False, logger=False).succeeded
 
     def find_wheel(self, folder, fatal=True):
         """list[str]: Wheel for this package found in 'folder', if any"""
         if runez.DRYRUN:
-            return ["%s-%s-py3-none-any.whl" % (self.wheelified, self.version)]
+            return [f"{self.wheelified}-{self.desired_track.version}-py3-none-any.whl"]
 
         result = []
         if folder and os.path.isdir(folder):
-            prefix = "%s-" % self.wheelified
+            prefix = f"{self.wheelified}-"
             for fname in os.listdir(folder):
                 if fname.startswith(prefix):
                     result.append(os.path.join(folder, fname))
@@ -312,33 +349,23 @@ class PackageSpec(object):
             if len(result) == 1:
                 return result[0]
 
-        return runez.abort("Expecting 1 wheel, found: %s" % (result or "None"), fatal=fatal, return_value=None)
-
-    def get_manifest(self):
-        """TrackedManifest: Manifest of the current installation of this package"""
-        manifest = TrackedManifest.from_file(self.manifest_path)
-        if not manifest:
-            # Temporary: take into account old v1 installs as well
-            old_base = self.cfg.meta.full_path(self.dashed)
-            old_manifest = runez.read_json(os.path.join(old_base, ".current.json"), default=None)
-            entry_points = runez.read_json(os.path.join(old_base, ".entry-points.json"), default=None)
-            if old_manifest and entry_points:
-                manifest = TrackedManifest(self.manifest_path, self.settings, entry_points)
-
-        return manifest
+        return runez.abort(f"Expecting 1 wheel, found: {runez.short(result)}", fatal=fatal, return_value=None)
 
     def groom_installation(self, keep_for=60):
         """
         Args:
             keep_for (int): Minimum time in minutes for how long to keep the previous latest version
         """
+        if self._pickley_dev_mode:
+            return
+
         if self.cfg.cache.contains(self.folder):
             runez.delete(self.folder, logger=False)
 
-        current = self.get_manifest()
         meta_path = self.meta_path
         current_age = None
-        if current and os.path.isdir(meta_path):
+        manifest = self.manifest
+        if manifest and os.path.isdir(meta_path):
             now = time.time()
             candidates = []
             for fname in os.listdir(meta_path):
@@ -348,7 +375,7 @@ class PackageSpec(object):
                 fpath = os.path.join(meta_path, fname)
                 vpart = fname[len(self.dashed) + 1:]
                 age = now - os.path.getmtime(fpath)
-                if vpart == current.version:
+                if vpart == manifest.version:
                     current_age = age
 
                 else:
@@ -380,65 +407,39 @@ class PackageSpec(object):
             self.settings,
             entry_points,
             pinned=self.pinned,
-            version=self.version,
+            version=self.desired_track.version,
         )
         payload = manifest.to_dict()
         runez.save_json(payload, self.manifest_path)
-        runez.save_json(payload, os.path.join(self.install_path, ".manifest.json"))
+        path = self.get_install_path(self.desired_track.version)
+        runez.save_json(payload, os.path.join(path, ".manifest.json"))
+        self._manifest = manifest
         return manifest
 
-    def get_desired_version_info(self, force=False):
-        """
-        Args:
-            force (bool): If True, ignore configured 'version_check_delay'
-
-        Returns:
-            (TrackedVersion): Object describing desired version
-        """
-        if self.version:
-            return TrackedVersion(source="explicit", version=self.version)
-
-        if self.pinned:
-            return TrackedVersion(source="pinned", version=self.pinned)
-
-        desired = self.get_latest(force=force)  # By default, the latest is desired
-        candidates = []
-        manifest = self.get_manifest()
-        if manifest and manifest.version:
-            candidates.append(TrackedVersion.from_manifest(manifest))
-
-        if self.dashed == PICKLEY:
-            candidates.append(TrackedVersion(source="current", version=__version__))
-
-        for candidate in candidates:
-            if candidate and candidate.version:
-                if not desired or not desired.version or Version(candidate.version) > Version(desired.version):
-                    desired = candidate
-
-        return desired
-
     def get_latest(self, force=False):
         """Tracked in DOT_META/.cache/<package>.latest"""
-        path = self.cfg.cache.full_path("%s.latest" % self.dashed)
-        age = self.cfg.version_check_delay(self)
-        if not force and age and runez.file.is_younger(path, age):
-            latest = TrackedVersion.from_file(path)
-            if latest:
-                return latest
+        if force or self._latest is None:
+            path = self.cfg.cache.full_path(f"{self.dashed}.latest")
+            if not force:
+                age = self.cfg.version_check_delay(self)
+                if age and runez.file.is_younger(path, age):
+                    self._latest = TrackedVersion.from_file(path)
+                    if self._latest:
+                        return self._latest
 
-        index = self.index
-        info = PypiInfo(index, self)
-        latest = TrackedVersion(index=index, problem=info.problem, source="latest", version=info.latest)
-        if not latest.problem:
-            runez.save_json(latest.to_dict(), path, fatal=None)
+            self._latest = TrackedVersion.from_pypi(self)
+            if not self._latest.problem:
+                runez.save_json(self._latest.to_dict(), path, fatal=None)
 
-        return latest
+            self._desired_track = None
+
+        return self._latest
 
 
 def get_default_index(*paths):
     """Configured pypi index from pip.conf"""
     for path in paths:
-        conf = runez.file.ini_to_dict(path, default={})
+        conf = runez.file.ini_to_dict(path)
         index = conf.get("global", {}).get("index-url")
         if index:
             return path, index
@@ -451,7 +452,7 @@ def get_program_path(path=None):
         path = runez.resolved_path(sys.argv[0])
 
     if path.endswith(".py") or path.endswith(".pyc"):
-        packaged = os.path.join(sys.prefix, "bin", "pickley")
+        packaged = runez.SYS_INFO.venv_bin_path(PICKLEY)
         if runez.is_executable(packaged):
             path = packaged  # Convenience when running from debugger
 
@@ -468,35 +469,36 @@ class PickleyConfig(object):
     configs = None  # type: list
     program_path = get_program_path()
 
+    _pickley_dev_path = None
+
     def __init__(self):
         self.configs = []
         self.config_path = None
         self.pip_conf, self.pip_conf_index = get_default_index("~/.config/pip/pip.conf", "/etc/pip.conf")
         self.default_index = self.pip_conf_index or DEFAULT_PYPI
         self._explored = set()
-        self._bundled_virtualenv_path = runez.UNSET
 
     def __repr__(self):
         return "<not-configured>" if self.base is None else runez.short(self.base)
 
-    @property
-    def bundled_virtualenv_path(self):
-        """str: Path to bundled virtualenv executable, if present"""
-        if self._bundled_virtualenv_path is runez.UNSET:
-            self._bundled_virtualenv_path = None
-            if sys.prefix != getattr(sys, "base_prefix", sys.prefix):
-                # We're running from a virtual environment
-                virtualenv = os.path.join(os.path.dirname(self.program_path), "virtualenv")
-                if runez.is_executable(virtualenv):
-                    self._bundled_virtualenv_path = virtualenv
-
-        return self._bundled_virtualenv_path
-
     @runez.cached_property
     def available_pythons(self):
         pyenv = self.pyenv()
-        scanner = pyenv_scanner(pyenv) if pyenv else None
-        return PythonDepot(scanner=scanner)
+        scanner = PythonInstallationScanner(pyenv) if pyenv else None
+        depot = PythonDepot(scanner=scanner)
+        depot.find_preferred_python(
+            self.get_value("preferred_pythons"),
+            min_python=self.get_value("min_python"),
+            preferred_min_python=self.get_value("preferred_min_python")
+        )
+        return depot
+
+    @classmethod
+    def pickley_dev_path(cls):
+        if cls._pickley_dev_path is None:
+            cls._pickley_dev_path = runez.DEV.project_folder
+
+        return cls._pickley_dev_path
 
     def set_base(self, base_path):
         """
@@ -513,10 +515,17 @@ class PickleyConfig(object):
 
         self._add_config_file(self.config_path)
         self._add_config_file(self.meta.full_path("config.json"))
-        defaults = dict(delivery="wrap", install_timeout=1800, python=DEFAULT_PYTHONS, version_check_delay=300)
+        defaults = dict(
+            delivery="wrap",
+            install_timeout=1800,
+            min_python="3.6",
+            preferred_min_python="3.7",
+            preferred_pythons="/usr/bin/python3,/usr/bin/python",
+            version_check_delay=300
+        )
         self.configs.append(RawConfig(self, "defaults", defaults))
 
-    def set_cli(self, config_path, delivery, index, python):
+    def set_cli(self, config_path, delivery, index, python, virtualenv):
         """
         Args:
             config_path (str | None): Optional configuration to use
@@ -525,12 +534,12 @@ class PickleyConfig(object):
             python (str | None): Optional python interpreter to use
         """
         self.config_path = config_path
-        self.cli = TrackedSettings(delivery, index, python)
+        self.cli = TrackedSettings(delivery, index, python, virtualenv)
 
     def _add_config_file(self, path, base=None):
         path = runez.resolved_path(path, base=base)
         if path and all(c.source != path for c in self.configs) and os.path.exists(path):
-            values = runez.read_json(path)
+            values = runez.read_json(path, logger=LOG.warning)
             if values:
                 self.configs.append(RawConfig(self, path, values))
                 included = values.get("include")
@@ -562,20 +571,19 @@ class PickleyConfig(object):
             (runez.pyenv.PythonInstallation): Object representing python installation
         """
         desired = self.get_value("python", pspec=pspec)
-        desired = [d.strip() for d in runez.flattened(desired, keep_empty=None, split=",")]
-        desired = [d for d in desired if d]
         if not desired:
-            # Edge case: configured empty python... just use invoker in that case
-            return self.available_pythons.invoker
+            # Most common case: use configured preferred python (will be 'invoker' by default)
+            return self.available_pythons.find_python(None)
 
         issues = []
         python = None
+        desired = runez.flattened(desired, split=",")
         for d in desired:
             python = self.available_pythons.find_python(d)
             if not python.problem:
                 return python
 
-            issues.append("Python '%s' skipped: %s" % (runez.bold(runez.short(d)), runez.red(python.problem)))
+            issues.append(f"Python '{runez.bold(runez.short(d))}' skipped: {runez.red(python.problem)}")
 
         for i in issues:  # Warn only if no python could be found at all
             LOG.warning(i)
@@ -585,7 +593,7 @@ class PickleyConfig(object):
 
         return python
 
-    def package_specs(self, names=None):
+    def package_specs(self, names=None, include_pickley=False):
         """
         Args:
             names (list | None): Package names, if empty: all installed
@@ -594,16 +602,21 @@ class PickleyConfig(object):
             (list[PackageSpec]): Corresponding PackageSpec-s
         """
         if names:
-            result = [self.resolved_bundle(name) for name in runez.flattened(names, split=" ")]
-            return [PackageSpec(self, name) for name in runez.flattened(result, unique=True)]
+            names = runez.flattened(names, split=" ")
+            if include_pickley and PICKLEY not in names:
+                names.append(PICKLEY)
+
+            result = [self.resolved_bundle(name) for name in names]
+            result = runez.flattened(result, unique=True)
+            return [PackageSpec(self, name) for name in result]
 
         result = []
         if os.path.isdir(self.meta.path):
             for fname in sorted(os.listdir(self.meta.path)):
-                if fname != PICKLEY:
+                if include_pickley or fname != PICKLEY:
                     fpath = os.path.join(self.meta.path, fname)
                     if os.path.isdir(fpath):
-                        if os.path.exists(os.path.join(fpath, ".manifest.json")) or os.path.exists(os.path.join(fpath, ".current.json")):
+                        if os.path.exists(os.path.join(fpath, ".manifest.json")):
                             result.append(PackageSpec(self, fname))
 
         return result
@@ -690,7 +703,7 @@ class PickleyConfig(object):
             if isinstance(pinned, dict):
                 return pinned.get("version")
 
-            if isinstance(pinned, runez.system.string_type):
+            if isinstance(pinned, str):
                 return pinned
 
     def pyenv(self):
@@ -722,6 +735,16 @@ class PickleyConfig(object):
         """
         return self.get_value("version_check_delay", pspec=pspec, validator=runez.to_int)
 
+    def get_virtualenv(self, pspec):
+        """
+        Args:
+            pspec (PackageSpec | None): Package spec, when applicable
+
+        Returns:
+            (str): Virtualenv version to use for this package spec (default: stdlib venv module)
+        """
+        return self.get_value("virtualenv", pspec=pspec)
+
     @staticmethod
     def colored_key(key, indent):
         if (key in K_CLI or key in K_LEAVES) and indent in (1, 3):
@@ -740,7 +763,7 @@ class PickleyConfig(object):
 
     def represented(self):
         """str: Human readable representation of this configuration"""
-        result = ["%s: %s" % (runez.bold("base"), self), ""]
+        result = [f"{runez.bold('base')}: {self}", ""]
         for c in self.configs:
             result.append(c.represented())
 
@@ -764,7 +787,25 @@ class TrackedVersion(object):
         self.version = version
 
     def __repr__(self):
-        return "%s (%s) %s" % (self.version, self.source, self.problem or "")
+        return self.version
+
+    @classmethod
+    def from_pypi(cls, pspec, index=None, include_prerelease=False):
+        """
+        Args:
+            pspec (PackageSpec): Pypi package name to lookup
+            index (str | None): URL to pypi index to use (default: pypi.org)
+            include_prerelease (bool): If True, include latest pre-release
+
+        Returns:
+            (TrackedVersion):
+        """
+        index = index or pspec.index or pspec.cfg.default_index
+        version = PypiStd.latest_pypi_version(pspec.dashed, index=index, include_prerelease=include_prerelease)
+        if not version:
+            return cls(index=index, problem=f"does not exist on {index}")
+
+        return cls(index=index, source="latest", version=version.text)
 
     @classmethod
     def from_manifest(cls, manifest, source="installed"):
@@ -772,7 +813,7 @@ class TrackedVersion(object):
 
     @classmethod
     def from_file(cls, path):
-        data = runez.read_json(path, default=None)
+        data = runez.read_json(path)
         if data:
             return cls(
                 index=data.get("index"),
@@ -805,17 +846,17 @@ class TrackedManifest(object):
     def __init__(self, path, settings, entrypoints, install_info=None, pinned=None, version=None):
         self.path = path
         self.settings = settings
-        self.entrypoints = entrypoints
+        self.entrypoints = entrypoints or {}
         self.install_info = install_info or TrackedInstallInfo.current()
         self.pinned = pinned
         self.version = version
 
     def __repr__(self):
-        return "%s [p: %s]" % (self.version, self.python)
+        return f"{self.version} [p: {self.python}]"
 
     @classmethod
     def from_file(cls, path):
-        data = runez.read_json(path, default=None)
+        data = runez.read_json(path)
         if data:
             return cls(
                 path,
@@ -885,11 +926,13 @@ class TrackedSettings(object):
     delivery = None  # type: str # Delivery method name
     index = None  # type: str # Pypi url used
     python = None  # type: str # Desired python
+    virtualenv = None  # type: str # Desired virtualenv version
 
-    def __init__(self, delivery, index, python):
+    def __init__(self, delivery, index, python, virtualenv):
         self.delivery = delivery
         self.index = index
         self.python = runez.short(python) if python else None
+        self.virtualenv = virtualenv
 
     @classmethod
     def from_manifest_data(cls, data):
@@ -899,10 +942,10 @@ class TrackedSettings(object):
     @classmethod
     def from_dict(cls, data):
         if data:
-            return cls(delivery=data.get("delivery"), index=data.get("index"), python=data.get("python"))
+            return cls(delivery=data.get("delivery"), index=data.get("index"), python=data.get("python"), virtualenv=data.get("virtualenv"))
 
     def to_dict(self):
-        return dict(delivery=self.delivery, index=self.index, python=self.python)
+        return dict(delivery=self.delivery, index=self.index, python=self.python, virtualenv=self.virtualenv)
 
 
 class FolderBase(object):
@@ -967,7 +1010,7 @@ class RawConfig(object):
         self.values = values
 
     def __repr__(self):
-        return "%s (%s values)" % (runez.short(self.source), len(self.values))
+        return f"{runez.short(self.source)} ({runez.plural(self.values)})"
 
     def get_nested(self, section, key):
         """
@@ -1019,20 +1062,20 @@ class RawConfig(object):
         for key, value in sorted(data.items()):
             key = self.parent.colored_key(key, indent)
             if isinstance(value, dict):
-                result.append("%s%s:" % (padding, key))
+                result.append(f"{padding}{key}:")
                 self._add_dict_representation(result, value, indent=indent + 1)
 
             elif isinstance(value, list):
-                result.append("%s%s:" % (padding, key))
+                result.append(f"{padding}{key}:")
                 for item in value:
-                    result.append("%s- %s" % (padding + "  ", runez.short(item)))
+                    result.append(f"{padding} - {runez.short(item)}")
 
             else:
-                result.append("%s%s: %s" % (padding, key, runez.short(value)))
+                result.append(f"{padding}{key}: {runez.short(value)}")
 
     def represented(self):
         """str: Human readable representation of this configuration"""
-        result = ["%s:" % runez.bold(runez.short(self.source))]
+        result = [f"{runez.bold(runez.short(self.source))}:"]
         if self.values:
             self._add_dict_representation(result, self.values)
 
