@@ -9,12 +9,12 @@ import time
 
 import click
 import runez
+from runez.pyenv import Version
 from runez.render import PrettyTable
 
-from pickley import __version__, abort, DOT_META, inform, PackageSpec, PickleyConfig, PLATFORM
-from pickley.delivery import DeliveryMethod, PICKLEY
+from pickley import __version__, abort, despecced, DOT_META, inform, PackageSpec, PickleyConfig, specced
+from pickley.delivery import PICKLEY
 from pickley.package import PexPackager, PythonVenv, VenvPackager
-from pickley.v1upgrade import V1Status
 
 
 LOG = logging.getLogger(__name__)
@@ -22,20 +22,6 @@ PACKAGER = VenvPackager  # Packager to use for this run
 CFG = PickleyConfig()
 
 
-def protected_main():
-    try:
-        main()
-
-    except KeyboardInterrupt:
-        abort(runez.red("\n\nAborted\n"))
-
-    except SoftLockException as e:
-        abort(e)
-
-    except NotImplementedError as e:
-        abort(runez.stringified(e) or "Not implemented")
-
-
 def setup_audit_log(cfg=CFG):
     """Setup audit.log log file handler"""
     if runez.DRYRUN:
@@ -45,7 +31,9 @@ def setup_audit_log(cfg=CFG):
         return
 
     if not runez.log.file_handler:
-        runez.log.progress.start(message_color=runez.dim, spinner_color=runez.bold)
+        if runez.color.is_coloring():
+            runez.log.progress.start(message_color=runez.dim, spinner_color=runez.bold)
+
         log_path = cfg.meta.full_path("audit.log")
         runez.log.trace("Logging to %s", log_path)
         runez.ensure_folder(cfg.meta.path)
@@ -82,7 +70,7 @@ class SoftLock(object):
         self.invalid = invalid or self.give_up * 2
 
     def __repr__(self):
-        return "lock %s" % self.pspec
+        return f"lock {self.pspec}"
 
     def _locked_by(self):
         """
@@ -93,7 +81,7 @@ class SoftLock(object):
             return None  # Lock file does not exist or invalidation age reached
 
         pid = None
-        for line in runez.readlines(self.lock_path, default=[], errors="ignore"):
+        for line in runez.readlines(self.lock_path):
             if pid is not None:
                 return line  # 2nd line hold CLI args process was invoked with
 
@@ -112,29 +100,28 @@ class SoftLock(object):
             if time.time() >= cutoff:
                 lock = runez.bold(runez.short(self.lock_path))
                 holder_args = runez.bold(holder_args)
-                raise SoftLockException("Can't grab lock %s, giving up\nIt is being held by: pickley %s" % (lock, holder_args))
+                raise SoftLockException(f"Can't grab lock {lock}, giving up\nIt is being held by: pickley {holder_args}")
 
             time.sleep(1)
             holder_args = self._locked_by()
 
         # We got the soft lock
         if runez.DRYRUN:
-            print("Would acquire %s" % runez.short(self.lock_path))
+            print(f"Would acquire {runez.short(self.lock_path)}")
 
         else:
-            runez.log.trace("Acquired %s" % runez.short(self.lock_path))
+            runez.log.trace(f"Acquired {runez.short(self.lock_path)}")
 
-        runez.write(self.lock_path, "%s\n%s\n" % (os.getpid(), runez.quoted(sys.argv[1:])), logger=False)
-        self.pspec.resolve()
+        runez.write(self.lock_path, runez.joined(os.getpid(), runez.quoted(sys.argv[1:]), delimiter="\n"), logger=False)
         return self
 
     def __exit__(self, *_):
         """Release lock"""
         if runez.DRYRUN:
-            print("Would release %s" % runez.short(self.lock_path))
+            print(f"Would release {runez.short(self.lock_path)}")
 
         else:
-            runez.log.trace("Released %s" % runez.short(self.lock_path))
+            runez.log.trace(f"Released {runez.short(self.lock_path)}")
 
         if CFG.base:
             runez.Anchored.pop(CFG.base.path)
@@ -142,7 +129,7 @@ class SoftLock(object):
         runez.delete(self.lock_path, logger=False)
 
 
-def perform_install(pspec, is_upgrade=False, force=False, quiet=False):
+def perform_install(pspec, is_upgrade=False, force=False, quiet=False, no_binary=None):
     """
     Args:
         pspec (PackageSpec): Package spec to install
@@ -155,44 +142,40 @@ def perform_install(pspec, is_upgrade=False, force=False, quiet=False):
     """
     with SoftLock(pspec):
         started = time.time()
-        pspec.resolve()
         skip_reason = pspec.skip_reason(force)
         if skip_reason:
-            inform("Skipping installation of %s: %s" % (pspec.dashed, runez.bold(skip_reason)))
+            inform(f"Skipping installation of {pspec}: {runez.bold(skip_reason)}")
             return None
 
-        manifest = pspec.get_manifest()
-        if is_upgrade and not manifest and not quiet:
-            abort("'%s' is not installed" % runez.red(pspec))
+        if is_upgrade and not pspec.is_currently_installed and not quiet:
+            abort(f"'{runez.red(pspec)}' is not installed")
 
-        if not pspec.version:
-            desired = pspec.get_desired_version_info(force=force)
-            if desired.problem:
-                action = "upgrade" if is_upgrade else "install"
-                abort("Can't %s %s: %s" % (action, pspec, runez.red(desired.problem)))
+        if force:
+            pspec.get_latest(force=force)
 
-            pspec.version = desired.version
+        if pspec.desired_track.problem:
+            action = "upgrade" if is_upgrade else "install"
+            abort(f"Can't {action} {pspec}: {runez.red(pspec.desired_track.problem)}")
 
-        if not force and manifest and manifest.version == pspec.version and pspec.is_healthily_installed():
+        if not force and pspec.is_up_to_date:
             if not quiet:
                 status = "up-to-date" if is_upgrade else "installed"
-                inform("%s v%s is already %s" % (pspec.dashed, runez.bold(pspec.version), status))
+                inform(f"{pspec.dashed} v{runez.bold(pspec.desired_track.version)} is already {status}")
 
             pspec.groom_installation()
-            return manifest
+            return
 
         setup_audit_log()
-        manifest = PACKAGER.install(pspec)
+        manifest = PACKAGER.install(pspec, no_binary=no_binary)
         if manifest and not quiet:
-            note = " in %s" % runez.represented_duration(time.time() - started)
+            note = f" in {runez.represented_duration(time.time() - started)}"
             action = "Upgraded" if is_upgrade else "Installed"
             if runez.DRYRUN:
-                action = "Would state: %s" % action
+                action = f"Would state: {action}"
 
-            inform("%s %s v%s%s" % (action, pspec.dashed, runez.bold(pspec.version), runez.dim(note)))
+            inform(f"{action} {pspec.dashed} v{runez.bold(pspec.desired_track.version)}{runez.dim(note)}")
 
         pspec.groom_installation()
-        return manifest
 
 
 def _find_base_from_program_path(path):
@@ -212,10 +195,10 @@ def _find_base_from_program_path(path):
 
 
 def find_base():
-    base_path = os.environ.get("PICKLEY_ROOT")
+    base_path = runez.resolved_path(os.environ.get("PICKLEY_ROOT"))
     if base_path:
         if not os.path.isdir(base_path):
-            abort("PICKLEY_ROOT points to non-existing directory %s" % runez.red(base_path))
+            abort(f"PICKLEY_ROOT points to non-existing directory {runez.red(base_path)}")
 
         return runez.resolved_path(base_path)
 
@@ -227,13 +210,14 @@ def clean_env_vars(*keys):
     """Ensure given env vars are removed if present"""
     for key in keys:
         if key in os.environ:
+            runez.log.trace(f"Unsetting env var {key}")
             del os.environ[key]
 
 
 @runez.click.group()
 @click.pass_context
 @runez.click.version(message="%(version)s", version=__version__)
-@runez.click.debug()
+@runez.click.debug("-v")
 @runez.click.dryrun("-n")
 @runez.click.color()
 @click.option("--config", "-c", metavar="PATH", help="Optional additional configuration to use")
@@ -241,25 +225,17 @@ def clean_env_vars(*keys):
 @click.option("--python", "-P", metavar="PATH", help="Python interpreter to use")
 @click.option("--delivery", "-d", help="Delivery method to use")
 @click.option("--packager", "-p", type=click.Choice(["pex", "venv"]), help="Packager to use")
-def main(ctx, debug, config, index, python, delivery, packager):
+@click.option("--virtualenv", help="Version of virtualenv to use (instead of built-in 'venv' module)")
+def main(ctx, debug, config, index, python, delivery, packager, virtualenv):
     """Package manager for python CLIs"""
     global PACKAGER
     PACKAGER = PexPackager if packager == "pex" else VenvPackager
     runez.system.AbortException = SystemExit
-    clean_env_vars("__PYVENV_LAUNCHER__", "PYTHONPATH")  # See https://github.com/python/cpython/pull/9516
-    if PLATFORM == "darwin" and "ARCHFLAGS" not in os.environ:
-        # Avoid issue on some OSX installations where ARM support seems to have been enabled too early
-        os.environ["ARCHFLAGS"] = "-arch x86_64"
-
     level = logging.WARNING
     if ctx.invoked_subcommand == "package":
         level = logging.INFO
         python = python or "invoker"  # Default to using invoker for 'package' subcommand
 
-    CFG.set_cli(config, delivery, index, python)
-    if ctx.invoked_subcommand != "package":
-        CFG.set_base(find_base())
-
     runez.log.setup(
         debug=debug or os.environ.get("PICKLEY_TRACE"),
         console_format="%(levelname)s %(message)s" if debug else "%(message)s",
@@ -268,125 +244,23 @@ def main(ctx, debug, config, index, python, delivery, packager):
         locations=None,
         trace="PICKLEY_TRACE",
     )
+    clean_env_vars("__PYVENV_LAUNCHER__", "PYTHONPATH")  # See https://github.com/python/cpython/pull/9516
+    if runez.SYS_INFO.platform_id.is_macos and "ARCHFLAGS" not in os.environ and runez.SYS_INFO.platform_id.arch:
+        # Ensure the proper platform is used on macos
+        archflags = f"-arch {runez.SYS_INFO.platform_id.arch}"
+        runez.log.trace(f"Setting ARCHFLAGS={archflags}")
+        os.environ["ARCHFLAGS"] = archflags
 
-
-def auto_upgrade_v1(cfg):
-    """Look for v1 installations, and upgrade them to v2"""
-    v1 = V1Status(cfg)
-    if v1.installed:
-        # On first auto-upgrade pickley (ran in background by wrapper)
-        setup_audit_log(cfg)
-        inform("Auto-upgrading %s packages with pickley v2" % len(v1.installed))
-        for prev in v1.installed:
-            pspec = PackageSpec(cfg, prev.name)
-            try:
-                manifest = perform_install(pspec, is_upgrade=False, quiet=False)
-                if manifest and manifest.entrypoints and prev.entrypoints:
-                    for old_ep in prev.entrypoints:
-                        if old_ep not in manifest.entrypoints:
-                            runez.delete(os.path.join(cfg.base.path, old_ep))
-
-            except BaseException:
-                inform("%s could not be upgraded, please reinstall it" % runez.red(prev.name))
-                if prev.entrypoints:
-                    for old_ep in prev.entrypoints:
-                        runez.delete(os.path.join(cfg.base.path, old_ep))
-
-            inform("----")
-
-        v1.clean_old_files()
-        inform("Done")
-        sys.exit(0)
-
-
-def _location_grand_parent():
-    return runez.parent_folder(runez.parent_folder(__file__))
-
-
-def needs_bootstrap(pickleyspec=None):
-    if pickleyspec is None:
-        grand_parent = _location_grand_parent()
-        if grand_parent and ".whl" in grand_parent:
-            return "Unpacking pex %s" % grand_parent
-
-        if "pickley.bootstrap" in sys.argv[0]:
-            return "Upgrading pass1"
-
-        return None
-
-    if pickleyspec.python and not pickleyspec.python.problem and CFG.available_pythons.invoker.major < pickleyspec.python.major:
-        return "Better python version available (%s vs %s)" % (CFG.available_pythons.invoker, pickleyspec.python)
-
-    manifest = pickleyspec.get_manifest()
-    if not manifest or not manifest.version:
-        return "Upgrading old pickley v1"  # We're running from an old pickley v1 install (no v2 manifest)
-
-
-def bootstrap():
-    """Bootstrap pickley (reinstall with venv instead of downloaded pex package)"""
-    pickleyspec = PackageSpec(CFG, PICKLEY, __version__)
-    grand_parent = _location_grand_parent()
-    if grand_parent and grand_parent.endswith(".whl"):
-        # We are running from a pex package
-        setup_audit_log()
-        LOG.debug("Bootstrapping pickley %s with %s" % (pickleyspec.version, pickleyspec.python))
-        venv = PythonVenv(pickleyspec)
-        with runez.TempFolder():
-            venv.run_python("-mwheel", "pack", grand_parent)
-            names = os.listdir(".")
-            if len(names) == 1:
-                venv.pip_install(names[0])
-                venv.run_python("-mpip", "install", "requests", "virtualenv", fatal=False)  # Inject extra packages to help bootstrap
-
-            else:
-                abort("Internal error: wheel pack %s" % grand_parent)
-
-        delivery = DeliveryMethod.delivery_method_by_name(pickleyspec.settings.delivery)
-        delivery.ping = False
-        delivery.install(pickleyspec, venv, {PICKLEY: "bootstrapped"})
-        runez.run(pickleyspec.exe_path(PICKLEY), "auto-upgrade", fatal=None, stdout=None, stderr=None)  # Trigger auto_upgrade_v1()
-        sys.exit(0)
-
-    if "pickley.bootstrap" in sys.argv[0]:
-        # We're running from pass1 bootstrap
-        setup_audit_log()
-        with SoftLock(pickleyspec):
-            VenvPackager.install(pickleyspec, ping=False)
-
-        LOG.debug("Pass 2 bootstrap done")
-        runez.run(pickleyspec.exe_path(PICKLEY), "auto-upgrade", fatal=None, stdout=None, stderr=None)  # Trigger auto_upgrade_v1()
-        sys.exit(0)
-
-    reason = needs_bootstrap(pickleyspec)
-    if reason:
-        setup_audit_log()
-        LOG.debug("Bootstrapping pickley: %s" % reason)
-        with SoftLock(pickleyspec):
-            delivery = DeliveryMethod.delivery_method_by_name(pickleyspec.settings.delivery)
-            delivery.ping = False
-            venv = PythonVenv(pickleyspec, folder=pickleyspec.cfg.meta.full_path(PICKLEY, "pickley.bootstrap"))
-            venv.pip_install(pickleyspec.specced)
-            src = venv.bin_path(PICKLEY)
-            dest = pickleyspec.exe_path(PICKLEY)
-            delivery._install(pickleyspec, dest, src)
-
-        LOG.debug("Pass 1 bootstrap done")
-        runez.run(pickleyspec.exe_path(PICKLEY), "auto-upgrade", fatal=None, stdout=None, stderr=None)  # Trigger pass 2
-        sys.exit(0)
+    CFG.set_cli(config, delivery, index, python, virtualenv)
+    if ctx.invoked_subcommand != "package":
+        CFG.set_base(find_base())
 
 
 @main.command(name="auto-upgrade")
 @click.option("--force", is_flag=True, help="Force auto-upgrade check, even if recently checked")
-@click.argument("package", required=False)
+@click.argument("package", required=True)
 def auto_upgrade(force, package):
     """Background auto-upgrade command (called by wrapper)"""
-    if not package or needs_bootstrap():
-        # We may need to bootstrap, or auto-upgrade v1, covered by test_config:test_v1
-        bootstrap()
-        auto_upgrade_v1(CFG)
-        if not package:
-            sys.exit(0)
-
     pspec = PackageSpec(CFG, package)
     ping = pspec.ping_path
     if not force and runez.file.is_younger(ping, 5):  # 5 seconds cool down on version check to avoid bursts
@@ -407,6 +281,16 @@ def auto_upgrade(force, package):
 def base(what):
     """Show pickley base folder"""
     path = CFG.base.path
+    if what == "bootstrap-own-wrapper":
+        # Internal: called by bootstrap script
+        from pickley.delivery import DeliveryMethodWrap
+
+        pspec = PackageSpec(CFG, f"{PICKLEY}=={__version__}")
+        venv = PythonVenv(CFG.meta.full_path(f"{PICKLEY}-{__version__}"), pspec, create=False)
+        wrap = DeliveryMethodWrap()
+        wrap.install(venv, {PICKLEY: PICKLEY})
+        return
+
     if what:
         paths = {
             "audit": CFG.meta.full_path("audit.log"),
@@ -419,16 +303,15 @@ def base(what):
         path = paths.get(what)
         if not path:
             options = [runez.green(s) for s in sorted(paths)]
-            abort("Unknown base folder reference '%s', try one of: %s" % (runez.red(what), ", ".join(options)))
+            abort(f"Unknown base folder reference '{runez.red(what)}', try one of: {', '.join(options)}")
 
     print(path)
 
 
 @main.command()
 @click.option("--force", "-f", is_flag=True, help="Force check, even if checked recently")
-@click.option("--verbose", "-v", is_flag=True, help="Show more information")
 @click.argument("packages", nargs=-1, required=False)
-def check(force, verbose, packages):
+def check(force, packages):
     """Check whether specified packages need an upgrade"""
     code = 0
     packages = CFG.package_specs(packages)
@@ -439,28 +322,27 @@ def check(force, verbose, packages):
     for pspec in packages:
         skip_reason = pspec.skip_reason(force)
         if skip_reason:
-            print("%s: %s, %s" % (pspec.dashed, runez.bold("skipped"), runez.dim(skip_reason)))
+            print(f"{pspec}: {runez.bold('skipped')}, {runez.dim(skip_reason)}]")
             continue
 
-        desired = pspec.get_desired_version_info(force=force)
-        dv = runez.bold(desired and desired.version)
-        manifest = pspec.get_manifest()
-        if desired.problem:
-            msg = desired.problem
+        pspec.get_latest(force=force)
+        dv = pspec.desired_track.version
+        if pspec.desired_track.problem:
+            msg = pspec.desired_track.problem
             code = 1
 
-        elif not manifest or not manifest.version:
-            msg = "v%s is not installed" % dv
+        elif not pspec.is_currently_installed:
+            msg = f"{runez.bold(dv)} not installed"
             code = 1
 
-        elif manifest.version == desired.version:
-            msg = "v%s is installed" % dv
+        elif pspec.is_up_to_date:
+            msg = f"{dv} up-to-date"
 
         else:
-            action = "upgraded to" if desired.source == "latest" else "caught up to %s" % desired.source
-            msg = "v%s installed, can be %s v%s" % (runez.dim(manifest.version), action, dv)
+            msg = runez.dim(pspec.manifest.version) if pspec.is_healthily_installed() else "unhealthy"
+            msg = f"{runez.bold(dv)} (currently {msg})"
 
-        print("%s: %s" % (pspec.dashed, msg))
+        print(f"{pspec.dashed}: {msg}")
 
     sys.exit(code)
 
@@ -476,9 +358,9 @@ def _diagnostics():
     yield "base", CFG.base
     yield "desired python", desired
     python = CFG.find_python(fatal=False)
-    yield "selected python", python.representation()
+    yield "selected python", python
     if python is not CFG.available_pythons.invoker:
-        yield "invoker python", CFG.available_pythons.invoker.representation()
+        yield "invoker python", CFG.available_pythons.invoker
 
     yield "default index", CFG.default_index
     yield "pip.conf", CFG.pip_conf
@@ -487,44 +369,184 @@ def _diagnostics():
 @main.command()
 def diagnostics():
     """Show diagnostics info"""
-    print(PrettyTable.two_column_diagnostics(_diagnostics))
     CFG.available_pythons.scan_path_env_var()
-    print("\n%s" % CFG.available_pythons.representation())
+    print(PrettyTable.two_column_diagnostics(_diagnostics(), runez.SYS_INFO.diagnostics(), CFG.available_pythons.representation()))
 
 
 @main.command()
 @click.option("--force", "-f", is_flag=True, help="Force installation, even if already installed")
+@click.option("--no-binary", help="Passed-through to pip install")
 @click.argument("packages", nargs=-1, required=True)
-def install(force, packages):
+def install(force, no_binary, packages):
     """Install a package from pypi"""
     setup_audit_log()
-    for pspec in CFG.package_specs(packages):
-        perform_install(pspec, is_upgrade=False, force=force, quiet=False)
+    specs = CFG.package_specs(packages, include_pickley=packages and packages[0].startswith("bundle:"))
+    for pspec in specs:
+        perform_install(pspec, is_upgrade=False, force=force, quiet=False, no_binary=no_binary)
+
+
+class TabularReport:
+    """Reports that can be shown as table, csv or json"""
+
+    def __init__(self, columns, additional=None, border="github", verbose=False):
+        """
+        Args:
+            columns (str | list): Column headers
+            additional (str | list | None): Additional column headers (showed when verbose is True)
+            border (str): Tabel border to use
+            verbose (bool): If True, show additional columns as well
+        """
+        self.border = border
+        self.verbose = verbose
+        cols = (columns, additional) if verbose else columns
+        self.columns = runez.flattened(cols, split=",")
+        self.table = PrettyTable(self.columns, border=border)
+        self.mapped_values = []
+        self.values = []
+
+    @staticmethod
+    def _json_key(key):
+        return runez.joined(runez.words(key.lower()), delimiter="-")
+
+    def add_row(self, **kwargs):
+        values = [kwargs.get(n) for n in self.columns]
+        self.mapped_values.append({self._json_key(k): runez.uncolored(v) for k, v in kwargs.items() if k in self.columns})
+        self.values.append(values)
+        self.table.add_row(values)
+
+    def represented(self, format):
+        if format in ("csv", "tsv"):
+            delimiter = "\t" if format == "tsv" else ","
+            lines = [runez.joined(self.columns, delimiter=delimiter)]
+            lines.extend(runez.joined([runez.uncolored(x) for x in v], delimiter=delimiter) for v in self.values)
+            return runez.joined(lines, delimiter="\n")
+
+        if format == "json":
+            return runez.represented_json(self.mapped_values)
+
+        if format == "yaml":
+            lines = []
+            for value in self.mapped_values:
+                text = [f" {k}: {value[k]}" for k in sorted(value)]
+                text = runez.joined(text, delimiter="\n ")
+                lines.append(f"-{text}")
+
+            return runez.joined(lines, delimiter="\n")
+
+        return self.table.get_string()
 
 
 @main.command()
 @runez.click.border("-b", default="github")
+@click.option("--format", "-f", type=click.Choice(["csv", "json", "tsv", "yaml"]), help="Representation format")
 @click.option("--verbose", "-v", is_flag=True, help="Show more information")
-def list(border, verbose):
+def list(border, format, verbose):
     """List installed packages"""
-    packages = CFG.package_specs()
+    packages = CFG.package_specs(include_pickley=verbose)
     if not packages:
         print("No packages installed")
         sys.exit(0)
 
-    table = PrettyTable("Package,Version,Delivery,Python,From index", border=border)
-    table.header.style = runez.bold
-    if not verbose:
-        table.header.hide(2, 3, 4)
-
+    report = TabularReport("Package,Version,Python", additional="Delivery,Index", border=border, verbose=verbose)
     for pspec in packages:
-        manifest = pspec.get_manifest()
+        manifest = pspec.manifest
         if manifest:
             python = CFG.available_pythons.find_python(manifest.python)
-            python = manifest.python if python.problem else python.representation()
-            table.add_row(pspec.dashed, manifest.version, manifest.delivery, python, manifest.index)
+            report.add_row(
+                Package=pspec.dashed,
+                Version=manifest.version,
+                Python=python,
+                Delivery=manifest.delivery,
+                Index=manifest.index
+            )
 
-    print(table)
+    print(report.represented(format))
+
+
+class RunSetup:
+    """Convenience defaults to use for 'run' commands"""
+
+    def __init__(self, command, package=None, pinned=None):
+        """
+        Args:
+            command (str): Name of command to run
+            package (str | None): Pypi package name to auto-install in venv (default: same as `command`)
+            pinned (str | None): Pinned version to use (implies per-project install)
+        """
+        self.command = command
+        self.package = package
+        self.pinned = pinned
+
+    def __repr__(self):
+        return self.canonical
+
+    @property
+    def canonical(self):
+        return self.specced if self.package == self.command else f"{self.specced}:{self.command}"
+
+    @property
+    def specced(self):
+        return specced(self.package, self.pinned)
+
+    @classmethod
+    def from_cli(cls, command):
+        package = command
+        if ":" in command:
+            # Allows to support cases where command name is different from pypi package name, ex: awscli:aws
+            package, _, command = command.rpartition(":")
+
+        package, pinned_package = despecced(package)
+        command, pinned_command = despecced(command)
+        pinned = pinned_package or pinned_command
+        rs = getattr(cls, "cmd_" + command.replace("-", "_"), None)
+        if rs:
+            rs = rs()
+            if pinned:
+                rs.pinned = pinned
+
+            return rs
+
+        return cls(command, package=package, pinned=pinned)
+
+    @classmethod
+    def perform_run(cls, command, args):
+        rs = cls.from_cli(command)
+        pspec = PackageSpec(CFG, rs.specced)
+        if not pspec.is_currently_installed:
+            perform_install(pspec, quiet=True)
+
+        runez.log.progress.stop()
+        path = pspec.exe_path(rs.command)
+        r = runez.run(path, args, stdout=None, stderr=None, fatal=False)
+        sys.exit(r.exit_code)
+
+    @classmethod
+    def cmd_aws(cls):
+        return cls("aws", package="awscli")
+
+    @classmethod
+    def cmd_pip_compile(cls):
+        return cls("pip-compile", package="pip-tools")
+
+
+@main.command(add_help_option=False, context_settings=dict(ignore_unknown_options=True))
+@click.argument("command")
+@click.argument("args", nargs=-1, type=click.UNPROCESSED)
+def run(command, args):
+    """
+    Run a python CLI (auto-install it if needed)
+
+    \b
+    Examples:
+         pickley run black
+         pickley run pip-compile
+         pickley run flake8 src/
+    """
+    if command == "--help":
+        click.echo(click.get_current_context().get_help())
+        return
+
+    RunSetup.perform_run(command, args)
 
 
 @main.command()
@@ -557,22 +579,74 @@ def uninstall(all, packages):
 
     setup_audit_log()
     for pspec in CFG.package_specs(packages):
-        manifest = pspec.get_manifest()
-        if not manifest or not manifest.version:
-            abort("%s was not installed with pickley" % runez.bold(pspec.dashed))
+        if not pspec.is_already_installed_by_pickley:
+            abort(f"{runez.bold(pspec.dashed)} was not installed with pickley")
 
-        if manifest.entrypoints:
-            for ep in manifest.entrypoints:
-                runez.delete(pspec.exe_path(ep))
+        for ep in pspec.manifest.entrypoints:
+            runez.delete(pspec.exe_path(ep))
 
         runez.delete(pspec.meta_path)
         action = "Would uninstall" if runez.DRYRUN else "Uninstalled"
-        inform("%s %s" % (action, pspec.dashed))
+        inform(f"{action} {pspec}")
 
     if all:
         runez.delete(CFG.base.full_path(PICKLEY))
         runez.delete(CFG.meta.path)
-        inform("pickley is now %s" % runez.red("uninstalled"))
+        inform(f"pickley is now {runez.red('uninstalled')}")
+
+
+@main.command()
+@click.option("--system", is_flag=True, help="Look at system PATH (not just pickley installs)")
+@click.argument("programs", nargs=-1)
+def version_check(system, programs):
+    """Check that programs are present with a minimum version"""
+    if not programs:
+        runez.abort("Specify at least one program to check")
+
+    specs = []
+    for program_spec in programs:
+        program, _, min_version = program_spec.partition(":")
+        min_version = Version(min_version)
+        if not program or not min_version.is_valid:
+            runez.abort(f"Invalid argument '{program_spec}', expecting format <program>:<version>")
+
+        specs.append((program, min_version))
+
+    overview = []
+    for program, min_version in specs:
+        if runez.DRYRUN:
+            runez.run(program, "--version")
+            continue
+
+        if system:
+            full_path = runez.which(program)
+            if not full_path:
+                runez.abort(f"{program} is not installed")
+
+        else:
+            pspec = PackageSpec(CFG, program)
+            full_path = pspec.exe_path(program)
+
+        r = runez.run(full_path, "--version", fatal=False, logger=None)
+        if not r.succeeded:
+            runez.abort(f"{runez.short(full_path)} --version failed: {runez.short(r.full_output)}")
+
+        version = parsed_version(r.full_output)
+        if not version or version < min_version:
+            runez.abort(f"{runez.short(full_path)} version too low: {version} (need {min_version}+)")
+
+        overview.append(f"{program} {version}")
+
+    print(runez.short(runez.joined(overview, delimiter=" ; ")))
+
+
+def parsed_version(text):
+    """Parse --version from text, in reverse order to avoid being fulled by warnings..."""
+    if text:
+        for line in reversed(text.splitlines()):
+            version = Version.from_text(line)
+            if version:
+                return version
 
 
 @main.command()
@@ -583,14 +657,14 @@ def uninstall(all, packages):
 @click.option("--sanity-check", default=None, show_default=True, help="Args to invoke produced package as a sanity check")
 @click.option("--requirement", "-r", multiple=True, help="Install from the given requirements file (can be used multiple times)")
 @click.argument("project", required=True)
-def package(build, dist, symlink, no_compile, sanity_check, project, requirement):
+@click.argument("additional", nargs=-1)
+def package(build, dist, symlink, no_compile, sanity_check, project, requirement, additional):
     """Package a project from source checkout"""
     started = time.time()
     runez.log.spec.default_logger = LOG.info
     CFG.set_base(runez.resolved_path(build))
-    finalizer = PackageFinalizer(project, dist, symlink)
+    finalizer = PackageFinalizer(project, dist, symlink, requirement, additional)
     finalizer.sanity_check = sanity_check
-    finalizer.requirements = requirement
     finalizer.compile = not no_compile
     finalizer.resolve()
     report = finalizer.finalize()
@@ -599,8 +673,8 @@ def package(build, dist, symlink, no_compile, sanity_check, project, requirement
         inform(report)
         inform("")
 
-    elapsed = "in %s" % runez.represented_duration(time.time() - started)
-    inform("Packaged %s successfully %s" % (runez.bold(runez.short(project)), runez.dim(elapsed)))
+    elapsed = f"in {runez.represented_duration(time.time() - started)}"
+    inform(f"Packaged {runez.bold(runez.short(project))} successfully {runez.dim(elapsed)}")
 
 
 class PackageFinalizer(object):
@@ -610,20 +684,30 @@ class PackageFinalizer(object):
 
     pspec = None  # type: PackageSpec
 
-    def __init__(self, project, dist, symlink):
+    def __init__(self, project, dist, symlink, requirement_files, additional, cfg=CFG):
         """
         Args:
             project (str): Folder where project to be packaged resides (must have a setup.py)
             dist (str): Relative path to folder to use as 'dist' (where to deliver package)
-            symlink (str | None): Synlink specification, of the form 'root:root/...'
+            symlink (str | None): Symlink specification, of the form 'root:root/...'
+            requirement_files (list | tuple): Requirement files to use
+            additional (list | tuple): Additional requirements
+            cfg (PickleyConfig): Config to use
         """
+        self.cfg = cfg
         self.folder = runez.resolved_path(project)
         self.dist = dist
         self.symlink = Symlinker(symlink) if symlink else None
         self.sanity_check = None
-        self.requirements = []
         self.compile = True
         self.border = "reddit"
+        if not requirement_files:
+            default_req = runez.resolved_path("requirements.txt", base=self.folder)
+            if os.path.exists(default_req):
+                requirement_files = default_req
+
+        requirement_files = [("-r", runez.resolved_path(r, base=self.folder)) for r in runez.flattened(requirement_files)]
+        self.requirements = runez.flattened(requirement_files, additional, self.folder, unique=True)
 
     @staticmethod
     def validate_sanity_check(exe, sanity_check):
@@ -633,30 +717,18 @@ class PackageFinalizer(object):
         r = runez.run(exe, sanity_check, fatal=False)
         if r.failed:
             if does_not_implement_cli_flag(r.output, r.error):
-                return "does not respond to %s" % sanity_check
+                return f"does not respond to {sanity_check}"
 
-            abort("'%s' failed %s sanity check: %s" % (exe, sanity_check, r.full_output))
+            abort(f"'{exe}' failed {sanity_check} sanity check: {r.full_output}")
 
         return runez.first_line(r.output or r.error)
 
     def resolve(self):
         if not os.path.isdir(self.folder):
-            abort("Folder %s does not exist" % runez.red(runez.short(self.folder)))
+            abort(f"Folder {runez.red(runez.short(self.folder))} does not exist")
 
-        req = self.requirements
-        if not req:
-            default_req = runez.resolved_path("requirements.txt", base=self.folder)
-            if os.path.exists(default_req):
-                req = [default_req]
-
-        if req:
-            req = [("-r", runez.resolved_path(r, base=self.folder)) for r in req]
-
-        req = runez.flattened(req, shellify=True)
-        req.append(self.folder)
-        self.requirements = req
-        self.pspec = PackageSpec(CFG, self.folder)
-        LOG.info("Using python: %s" % self.pspec.python)
+        self.pspec = PackageSpec(self.cfg, self.folder)
+        LOG.info(f"Using python: {self.pspec.python}")
         if self.dist.startswith("root/"):
             # Special case: we're targeting 'root/...' probably for a debian, use target in that case to avoid venv relocation issues
             target = self.dist[4:]
@@ -671,10 +743,10 @@ class PackageFinalizer(object):
 
     def finalize(self):
         """Run sanity check and/or symlinks, and return a report"""
-        with runez.Anchored(self.folder, CFG.base.path):
-            runez.ensure_folder(CFG.base.path, clean=True, logger=False)
+        with runez.Anchored(self.folder, self.cfg.base.path):
+            runez.ensure_folder(self.cfg.base.path, clean=True, logger=False)
             dist_folder = runez.resolved_path(self.dist)
-            exes = PACKAGER.package(self.pspec, CFG.base.path, dist_folder, self.requirements, self.compile)
+            exes = PACKAGER.package(self.pspec, self.cfg.base.path, dist_folder, self.requirements, self.compile)
             if exes:
                 report = PrettyTable(["Packaged executable", self.sanity_check], border=self.border)
                 report.header.style = "bold"
@@ -706,7 +778,7 @@ class Symlinker(object):
     def __init__(self, spec):
         self.base, _, self.target = spec.partition(":")
         if not self.base or not self.target:
-            abort("Invalid symlink specification '%s'" % spec)
+            abort(f"Invalid symlink specification '{spec}'")
 
         self.base = runez.resolved_path(self.base)
         self.target = runez.resolved_path(self.target)
@@ -714,13 +786,12 @@ class Symlinker(object):
     def apply(self, exe):
         dest = os.path.join(self.target, os.path.basename(exe))
         if os.path.exists(exe):
-            runez.delete(dest, logger=False)
             r = runez.symlink(exe, dest, must_exist=False)
             if r > 0:
-                inform("Symlinked %s -> %s" % (runez.short(dest), runez.short(exe)))
+                inform(f"Symlinked {runez.short(dest)} -> {runez.short(exe)}")
 
         else:
-            LOG.debug("'%s' does not exist, skipping symlink" % exe)
+            LOG.debug(f"'{exe}' does not exist, skipping symlink")
 
 
 def delete_file(path):
