@@ -3,21 +3,14 @@ import os
 
 import runez
 
-from pickley import abort, PICKLEY
+from pickley import abort, PackageSpec, PICKLEY
+from pickley.bstrap import create_virtualenv, pip_version
 from pickley.delivery import DeliveryMethod
 
 
 LOG = logging.getLogger(__name__)
 
 
-def download_command(target, url):
-    curl = runez.which("curl")
-    if curl:
-        return [curl, "-s", "-o", target, url]
-
-    return ["wget", "-q", "-O%s" % target, url]
-
-
 class PackageFolder(object):
     """Allows to track reported file contents by `pip show -f`"""
 
@@ -27,7 +20,7 @@ class PackageFolder(object):
         self.files = {}
 
     def __str__(self):
-        return "%s [%s files]" % (self.folder, len(self))
+        return f"{self.folder} [{len(self)} files]"
 
     def __len__(self):
         return len(self.files)
@@ -46,30 +39,33 @@ class PackageFolder(object):
 class PackageContents(object):
     """Contents of a pip-installed package"""
 
-    def __init__(self, venv, pspec):
+    def __init__(self, venv):
         """
         Args:
             venv (PythonVenv): Venv to extract info from
-            pspec (pickley.PackageSpec): Package spec to look for
         """
         self.venv = venv
-        self.pspec = pspec
         self.location = None
         self.bin = PackageFolder(folder="bin")
         self.completers = PackageFolder(folder="bin")
         self.dist_info = PackageFolder()
         self.files = None
         self.info = {}
-        name = pspec.dashed
+        name = venv.pspec.dashed
         if runez.DRYRUN and not venv.is_venv_exe("bin/pip"):
-            self.bin.files = {pspec.dashed: "dryrun"}  # Pretend an entry point exists in dryrun mode
+            self.bin.files = {venv.pspec.dashed: "dryrun"}  # Pretend an entry point exists in dryrun mode
+            return
+
+        if name == PICKLEY:
+            self.bin.files = {name: name}  # When pickley is installed with -e (--editable), metadata is not "standard"
             return
 
         if name == "ansible":
             # Is there a better way to detect weird indirections like ansible does?
-            name = "ansible-core" if not pspec.version or pspec.version >= "4" else "ansible-base"
+            version = venv.pspec.desired_track.version
+            name = "ansible-core" if not version or version >= "4" else "ansible-base"
 
-        r = venv.run_python("-mpip", "show", "-f", name, fatal=False)
+        r = venv.run_pip("show", "-f", name, fatal=False)
         if not r.succeeded:
             return
 
@@ -111,7 +107,7 @@ class PackageContents(object):
                     self.location = value
 
     def __repr__(self):
-        return "%s [%s]" % (self.pspec, runez.short(self.location))
+        return f"{self.venv.pspec} [{runez.short(self.location)}]"
 
     @runez.cached_property
     def entry_points(self):
@@ -124,80 +120,72 @@ class PackageContents(object):
                 if isinstance(commands, dict):
                     wrap_console = commands.get("wrap_console")
                     if wrap_console:
-                        runez.log.trace("Found %s entry points in metadata.json" % len(wrap_console))
+                        nb = runez.plural(wrap_console, "entry point")
+                        runez.log.trace(f"Found {nb} in metadata.json")
                         return wrap_console
 
         entry_points_txt = self.dist_info.files.get("entry_points.txt")
         if entry_points_txt:
-            metadata = runez.file.ini_to_dict(entry_points_txt, default={})
+            metadata = runez.file.ini_to_dict(entry_points_txt)
             console_scripts = metadata.get("console_scripts")
             if console_scripts:
-                runez.log.trace("Found %s entry points in entry_points.txt" % len(console_scripts))
+                nb = runez.plural(console_scripts, "entry point")
+                runez.log.trace(f"Found {nb} in entry_points.txt")
                 return console_scripts
 
         if self.bin.files:
-            runez.log.trace("Found %s bin/ scripts" % len(self.bin.files))
+            nb = runez.plural(self.bin.files, "bin/ script")
+            runez.log.trace(f"Found {nb}")
 
         return self.bin.files or None
 
 
 class PythonVenv(object):
-    def __init__(self, pspec=None, folder=None, python=None, index=None, cfg=None):
+
+    def __init__(self, folder, pspec, create=True):
         """
         Args:
-            pspec (pickley.PackageSpec | None): Package spec to install
-            folder (str | None): Target folder (default: pspec.install_path)
-            python (pickley.env.PythonInstallation): Python to use (default: pspec.python)
-            index (str | None): Optional custom pypi index to use (default: pspec.index)
-            cfg (pickley.PickleyConfig | None): Config to use
+            folder (str | pathlib.Path): Target folder
+            pspec (pickley.PackageSpec): Package spec to install
+            create (bool): Create venv if True
         """
-        if folder is None and pspec:
-            folder = pspec.install_path
-
-        if not python and pspec:
-            python = pspec.python
-
-        if not index and pspec:
-            index = pspec.index
-
         self.folder = folder
-        self.index = index
-        self.py_path = os.path.join(folder, "bin", "python")
-        if folder:
-            cfg = cfg or pspec.cfg
-            python = python or cfg.find_python(pspec=pspec)
-            runez.ensure_folder(folder, clean=True, logger=False)
-            if cfg.bundled_virtualenv_path:
-                runez.run(cfg.bundled_virtualenv_path, "-p", python.executable, folder)
-                return
-
-            if python.major > 2 and (not pspec or pspec.dashed != "tox"):
-                # See https://github.com/tox-dev/tox/issues/1689
-                r = runez.run(python.executable, "-mvenv", self.folder, fatal=False)
-                if r.succeeded:
-                    pip = "pip"
-                    if os.environ.get("VIRTUALENV_PIP"):  # pragma: no cover
-                        # Optionally respect https://tox.readthedocs.io/en/latest/config.html#conf-download
-                        pip += "==%s" % os.environ.get("VIRTUALENV_PIP")
-
-                    self.pip_install("-U", pip, "setuptools", "wheel")
-                    return
-
-                LOG.debug("Module venv failed, trying virtualenv bootstrap")  # pragma: no cover
-
-            runez.ensure_folder(cfg.cache.path)
-            zipapp = os.path.realpath(cfg.cache.full_path("virtualenv.pyz"))
-            if not runez.file.is_younger(zipapp, runez.date.SECONDS_IN_ONE_DAY):
-                runez.delete(zipapp, fatal=False)
-                args = download_command(zipapp, "https://bootstrap.pypa.io/virtualenv/virtualenv.pyz")
-                runez.run(*args)
-
-            runez.run(python.executable, zipapp, folder)
+        self.pspec = pspec
+        self.python = pspec.python or pspec.cfg.find_python(pspec)
+        if create:
+            self._create_virtualenv(vv=pspec.cfg.get_virtualenv(pspec))
 
     def __repr__(self):
         return runez.short(self.folder)
 
-    def bin_path(self, name):
+    def _create_virtualenv(self, runner=runez.run, vv=None):
+        runez.ensure_folder(self.folder, clean=True, logger=False)
+        if vv:
+            return self._old_virtualenv(runner, vv)
+
+        runez.run(self.python.executable, "-mvenv", self.folder)
+        if not runez.DRYRUN and not self.pip_path:
+            return self._old_virtualenv(runner, "latest")
+
+        pip_spec = pip_version(self.python.version.components)
+        pip_spec = f"pip=={pip_spec}" if pip_spec else "pip"
+        return self.run_pip("install", "-U", pip_spec)
+
+    def _old_virtualenv(self, runner, vv):
+        """Create a virtualenv using old virtualenv module"""
+        pv = self.python.version.components
+        if not vv or vv == "latest":
+            vv = PackageSpec(self.pspec.cfg, "virtualenv")
+            vv = vv.desired_track.version
+
+        tmp = self.pspec.cfg.cache.path
+        return create_virtualenv(tmp, pv, self.python.executable, self.folder, virtualenv_version=vv, runner=runner, dryrun=runez.DRYRUN)
+
+    @property
+    def pip_path(self):
+        return self.bin_path("pip", try_variant=True)
+
+    def bin_path(self, name, try_variant=False):
         """
         Args:
             name (str): File name
@@ -205,7 +193,14 @@ class PythonVenv(object):
         Returns:
             (str): Full path to this <venv>/bin/<name>
         """
-        return os.path.join(self.folder, "bin", name)
+        path = os.path.join(self.folder, "bin", name)
+        if runez.DRYRUN or os.path.exists(path):
+            return path
+
+        if try_variant:
+            path = os.path.join(self.folder, "bin", f"{name}3")
+            if os.path.exists(path):
+                return path
 
     def is_venv_exe(self, path):
         """
@@ -216,14 +211,14 @@ class PythonVenv(object):
             (bool): True if 'path' points to a python executable part of this venv
         """
         if runez.is_executable(path):
-            lines = runez.readlines(path, default=None, first=2, errors="ignore")
-            if lines and len(lines) > 1 and lines[0].startswith("#!"):
-                if self.folder in lines[0] or self.folder in lines[1]:  # 2 variants: "#!<folder>/bin/python" or 'exec "<folder>/bin/..."'
-                    return True
+            for line in runez.readlines(path, first=1):
+                if line.startswith("#!"):
+                    if path.startswith(self.folder):
+                        return True
 
     def pip_install(self, *args):
         """Allows to not forget to state the -i index..."""
-        r = self._run_pip("install", "-i", self.index, *args, fatal=False)
+        r = self.run_pip("install", "-i", self.pspec.index, *args, fatal=False)
         if r.failed:
             message = "\n".join(simplified_pip_error(r.error, r.output))
             abort(message)
@@ -232,14 +227,16 @@ class PythonVenv(object):
 
     def pip_wheel(self, *args):
         """Allows to not forget to state the -i index..."""
-        return self._run_pip("wheel", "-i", self.index, *args)
+        return self.run_pip("wheel", "-i", self.pspec.index, *args)
+
+    def run_pip(self, *args, **kwargs):
+        return runez.run(self.pip_path, *args, **kwargs)
 
     def run_python(self, *args, **kwargs):
         """Run python from this venv with given args"""
-        return runez.run(self.py_path, *args, **kwargs)
-
-    def _run_pip(self, *args, **kwargs):
-        return self.run_python("-mpip", "--isolated", *args, **kwargs)
+        # kwargs.setdefault("short_exe", True)
+        exe = self.bin_path("python", try_variant=True)
+        return runez.run(exe, *args, **kwargs)
 
 
 def simplified_pip_error(error, output):
@@ -254,7 +251,7 @@ class Packager(object):
     """Ancestor to package/install implementations"""
 
     @staticmethod
-    def install(pspec, ping=True):
+    def install(pspec, ping=True, no_binary=None):
         """
         Args:
             pspec (pickley.PackageSpec): Targeted package spec
@@ -282,24 +279,22 @@ class PexPackager(Packager):
     """Package via pex (https://pypi.org/project/pex/)"""
 
     @staticmethod
-    def install(pspec, ping=True):  # pragma: no cover
+    def install(pspec, ping=True, no_binary=None):
         raise NotImplementedError("Installation with 'PexPackager' is not supported")
 
     @staticmethod
     def package(pspec, build_folder, dist_folder, requirements, run_compile_all):
         runez.ensure_folder(build_folder, clean=True)
-        if pspec.python.major < 3:  # pragma: no cover
-            abort("Packaging with pex is not supported any more with python2")
-
+        runez.abort_if(pspec.python.major < 3, "Packaging with pex is not supported any more with python2")
         pex_root = os.path.join(build_folder, "pex-root")
         tmp = os.path.join(build_folder, "pex-tmp")
         wheels = os.path.join(build_folder, "wheels")
         runez.ensure_folder(tmp, logger=False)
         runez.ensure_folder(wheels, logger=False)
-        pex_venv = PythonVenv(pspec, folder=os.path.join(build_folder, "pex-venv"))
-        pex_venv.pip_install("pex==2.1.42", *requirements)
+        pex_venv = PythonVenv(os.path.join(build_folder, "pex-venv"), pspec)
+        pex_venv.pip_install("pex==2.1.90", *requirements)
         pex_venv.pip_wheel("--cache-dir", wheels, "--wheel-dir", wheels, *requirements)
-        contents = PackageContents(pex_venv, pspec)
+        contents = PackageContents(pex_venv)
         if contents.entry_points:
             wheel_path = pspec.find_wheel(wheels)
             result = []
@@ -307,11 +302,11 @@ class PexPackager(Packager):
                 target = os.path.join(dist_folder, name)
                 runez.delete(target)
                 pex_venv.run_python(
-                    "-mpex", "-o%s" % target, "--pex-root", pex_root, "--tmpdir", tmp,
+                    "-mpex", f"-o{target}", "--pex-root", pex_root, "--tmpdir", tmp,
                     "--no-index", "--find-links", wheels,  # resolver options
                     None if run_compile_all else "--no-compile",  # output options
-                    "-c%s" % name,  # entry point options
-                    "--python-shebang", "/usr/bin/env python%s" % pspec.python.major,
+                    f"-c{name}",  # entry point options
+                    "--python-shebang", f"/usr/bin/env python{pspec.python.major}",
                     wheel_path,
                 )
                 result.append(target)
@@ -323,42 +318,43 @@ class VenvPackager(Packager):
     """Install in a virtualenv"""
 
     @staticmethod
-    def install(pspec, ping=True):
+    def install(pspec, ping=True, no_binary=None):
         delivery = DeliveryMethod.delivery_method_by_name(pspec.settings.delivery)
         delivery.ping = ping
-        args = [pspec.specced]
-        extras = None
+        args = []
+        if no_binary:
+            args.append("--no-binary")
+            args.append(no_binary)
+
+        venv_folder = pspec.get_install_path(pspec.desired_track.version)
         if pspec.folder:
-            args = [pspec.folder]
+            args.append(pspec.folder)
 
-        elif pspec.dashed == PICKLEY:
-            # Inject extra packages for pickley, to help bootstrap
-            extras = ["virtualenv", "requests"]
-            project_path = runez.log.project_path()
-            if project_path:
-                args = ["-e", project_path]  # Development mode (running from source checkout)
+        elif pspec._pickley_dev_mode:
+            args.append("-e")
+            args.append(pspec._pickley_dev_mode)
 
-        venv = PythonVenv(pspec)
+        else:
+            args.append(f"{pspec.dashed}=={pspec.desired_track.version}")
+
+        venv = PythonVenv(venv_folder, pspec)
         venv.pip_install(*args)
-        if extras:
-            venv.pip_install(*extras)
-
-        contents = PackageContents(venv, pspec)
+        contents = PackageContents(venv)
         if not contents.entry_points:
             runez.delete(pspec.meta_path)
-            abort("Can't install '%s', it is %s" % (runez.bold(pspec.dashed), runez.red("not a CLI")))
+            abort(f"Can't install '{runez.bold(pspec.dashed)}', it is {runez.red('not a CLI')}")
 
-        return delivery.install(pspec, venv, contents.entry_points)
+        return delivery.install(venv, contents.entry_points)
 
     @staticmethod
     def package(pspec, build_folder, dist_folder, requirements, run_compile_all):
         runez.ensure_folder(dist_folder, clean=True, logger=False)
-        venv = PythonVenv(pspec, folder=dist_folder)
+        venv = PythonVenv(dist_folder, pspec)
         venv.pip_install(*requirements)
         if run_compile_all:
             venv.run_python("-mcompileall", dist_folder)
 
-        contents = PackageContents(venv, pspec)
+        contents = PackageContents(venv)
         if contents.entry_points:
             result = []
             for name in contents.entry_points:
